import os
import sys
import argparse
from collections import namedtuple
from monolith.compat import OrderedDict
from monolith.compat import unicode
from monolith.cli.exceptions import AlreadyRegistered


Argument = namedtuple('Argument', 'args kwargs')


def arg(*args, **kwargs):
    """
    Returns *Argument* namedtuple in format: ``(args, kwargs)``. In example::
    
        >>> arg(1, 2, 'foo', 'bar')
        Argument(args=(1, 2, 'foo', 'bar'), kwargs={})
        >>> arg('a', 1, foo='bar')
        Argument(args=('a', 1), kwargs={'foo': 'bar'})

    """
    return Argument(args, kwargs)


class Parser(argparse.ArgumentParser):
    """
    Subclass of ``argparse.ArgumentParser`` providing more control over output
    stream.
    """
    def __init__(self, *args, **kwargs):
        self.stream = kwargs.pop('stream', sys.stderr)
        super(Parser, self).__init__(*args, **kwargs)

    def _print_message(self, message, file=None):
        if file is None:
            file = self.stream
        super(Parser, self)._print_message(unicode(message), file)


class ExecutionManager(object):
    usage = None
    completion = False
    completion_env_var_name = ''
    parser_cls = Parser

    def __init__(self, argv=None, stderr=None, stdout=None):
        if argv is None:
            argv = [a for a in sys.argv]
        self.prog_name = os.path.basename(argv[0])
        self.argv = argv[1:]
        self.registry = {}
        self.stderr = stderr or sys.stderr
        self.stdout = stdout or sys.stdout

        for name, Command in self.get_commands_to_register().items():
            self.register(name, Command)

    def get_usage(self):
        """
        Returns *usage* text of the main application parser.
        """
        return self.usage

    def get_parser(self):
        """
        Returns :class:`monolith.cli.Parser` instance for this
        *ExecutionManager*.
        """
        parser = self.parser_cls(prog=self.prog_name, usage=self.get_usage(),
            stream=self.stderr)
        subparsers = parser.add_subparsers(
            title='subcommands',
        )
        for name, command in self.registry.items():
            cmdparser = subparsers.add_parser(name, help=command.help)
            for argument in command.get_args():
                cmdparser.add_argument(*argument.args, **argument.kwargs)
            cmdparser.set_defaults(func=command.handle)

        return parser

    def register(self, name, Command, force=False):
        """
        Registers given ``Command`` (as given ``name``) at this
        *ExecutionManager*'s registry.

        :param name: name in the registry under which given ``Command``
          should be stored.
        :param Command: should be subclass of
          :class:``monolith.cli.base.BaseCommand``
        :param force: Forces registration if set to ``True`` - even if another
          command was already registered, it would be overridden and no
          execption would be raised. Defaults to ``False``.

        :raises AlreadyRegistered: If another command was already registered
          under given ``name``.
        """
        if not force and name in self.registry:
            raise AlreadyRegistered('Command %r is already registered' % name)
        command = Command(self.prog_name, self.stdout)
        self.registry[name] = command
        command.post_register(self)

    def get_commands(self):
        """
        Returns commands stored in the registry (sorted by name).
        """
        commands = OrderedDict()
        for cmd in sorted(self.registry.keys()):
            commands[cmd] = self.registry[cmd]
        return commands

    def get_commands_to_register(self):
        """
        Returns dictionary (*name* / *Command* or string pointing at the
        command class.
        """
        return {}

    def call_command(self, cmd, *argv):
        """
        Runs a command.

        :param cmd: command to run (key at the registry)
        :param argv: arguments that would be passed to the command
        """
        parser = self.get_parser()
        args = [cmd] + list(argv)
        namespace = parser.parse_args(args)
        namespace.func(namespace)

    def execute(self, argv=None):
        """
        Executes command based on given arguments.
        """
        if self.completion:
            self.autocomplete()
        parser = self.get_parser()
        namespace = parser.parse_args(argv)
        if hasattr(namespace, 'func'):
            namespace.func(namespace)

    def autocomplete(self):
        """
        If *completion* is enabled, this method would write to ``self.stdout``
        completion words separated with space.
        """
        if self.completion_env_var_name not in os.environ:
            return
        cwords = os.environ['COMP_WORDS'].split()[1:]
        cword = int(os.environ['COMP_CWORD'])
        try:
            current = cwords[cword-1]
        except IndexError:
            current = ''
        cmd_names = self.get_commands().keys()

        if current:
            self.stdout.write(unicode(' '.join(
                [name for name in cmd_names if name.startswith(current)])))

        sys.exit(1)


class BaseCommand(object):
    """
    Base command class that should be subclassed by concrete commands.

    **Attributes**

    - ``help``: Help description for this command. Defaults to empty string.
    - ``args``: List of :class:`Argument` instances. Defaults to empty list.
    - ``prog_name``: Program name of *ExecutionManager* within which this
      command is run. Defaults to ``None``.
    - ``stdout``: File-like object. Command should write to it. Defaults to
      ``sys.stdout``.
    """
    help = ''
    args = []

    def __init__(self, prog_name=None, stdout=None):
        self.prog_name = prog_name or ''
        self.stdout = stdout or sys.stdout

    def get_args(self):
        """
        Returns list of :class:`Argument` instances for the parser. By default,
        it returns ``self.args``.
        """
        return self.args or []

    def handle(self, namespace):
        """
        Handles given ``namespace`` and executes command. Should be overridden
        at subclass.
        """
        raise NotImplementedError

    def post_register(self, manager):
        """
        Performs actions once this command is registered within given
        ``manager``. By default it does nothing.
        """
        pass


class LabelCommand(BaseCommand):
    """
    Command that works on given position arguments (*labels*). By default, at
    least one *label* is required. This is controlled by *labels_required*
    attribute.

    **Extra attributes**:

    - ``labels_required``: If ``True``, at least one *label* is required,
      otherwise no positional arguments could be given. Defaults to ``True``.
    """
    labels_required = True

    def get_labels_arg(self):
        """
        Returns argument for *labels*.
        """
        nargs = self.labels_required and '+' or '*'
        return arg('labels', nargs=nargs)

    def get_args(self):
        return self.args + [self.get_labels_arg()]

    def handle(self, namespace):
        """
        Handles given ``namespace`` by calling ``handle_label`` method
        for each given *label*.
        """
        for label in namespace.labels:
            self.handle_label(label, namespace)
        else:
            self.handle_no_labels(namespace)

    def handle_label(self, label, namespace):
        """
        Handles single *label*. Should be overridden at subclass.
        """
        raise NotImplementedError

    def handle_no_labels(self, namespace):
        """
        Performs some action if no *lables* were given. By default it does
        nothing.
        """
        pass


class SingleLabelCommand(BaseCommand):
    """
    Command that works on given positional argument (*label*).

    **Extra arguments**:

    - ``label_default_value``: If no *label* were given, this would be default
      value that would be passed to ``namespace``. Defaults to ``None``.
    """
    label_default_value = None

    def get_label_arg(self):
        """
        Returns argument for *label*.
        """
        return arg('label', default=self.label_default_value, nargs='?')

    def get_args(self):
        return self.args + [self.get_label_arg()]

    def handle(self, namespace):
        """
        Calls ``handle_label`` method for given *label*.
        """
        self.handle_label(namespace.label, namespace)

    def handle_label(self, label, namespace):
        """
        Handles *label*. Should be overridden at subclass.
        """
        raise NotImplementedError




from monolith.compat import unicode
from monolith.cli.base import BaseCommand


COMPLETION_TEMPLATE = '''
# %(prog_name)s bash completion start
_%(prog_name)s_completion()
{
    COMPREPLY=( $( COMP_WORDS="${COMP_WORDS[*]}" \\
                   COMP_CWORD=$COMP_CWORD \\
                   %(ENV_VAR_NAME)s=1 $1 ) )
}
complete -o default -F _%(prog_name)s_completion %(prog_name)s
# %(prog_name)s bash completion end

'''


class CompletionCommand(BaseCommand):
    help = ''.join((
        'Prints out shell snippet that once evaluated would allow '
        'this command utility to use completion abilities.',
    ))
    template = COMPLETION_TEMPLATE

    def get_env_var_name(self):
        return '_'.join((self.prog_name.upper(), 'AUTO_COMPLETE'))

    def get_completion_snippet(self):
        return self.template % {'prog_name': self.prog_name,
            'ENV_VAR_NAME': self.get_env_var_name()}

    def handle(self, namespace):
        self.stdout.write(unicode(self.get_completion_snippet()))

    def post_register(self, manager):
        manager.completion = True
        manager.completion_env_var_name = self.get_env_var_name()




from .base import BaseCommand
from .base import ExecutionManager
from .base import LabelCommand
from .base import Parser
from .base import SingleLabelCommand
from .base import arg
from .completion import CompletionCommand


__all__ = ['ExecutionManager', 'arg',
    'Parser',
    'BaseCommand',
    'LabelCommand', 'SingleLabelCommand',
    'CompletionCommand',
]




import io
import os
import sys
import mock
import argparse
from monolith.compat import unittest
from monolith.cli.base import arg
from monolith.cli.base import ExecutionManager
from monolith.cli.base import BaseCommand
from monolith.cli.completion import CompletionCommand


class TestCompletionCommand(unittest.TestCase):

    def test_get_env_var_name(self):
        command = CompletionCommand('foo')
        self.assertEqual(command.get_env_var_name(), 'FOO_AUTO_COMPLETE')

    def test_get_completion_snippet(self):
        command = CompletionCommand('foo')
        command.template = '%(prog_name)s | %(ENV_VAR_NAME)s'
        self.assertEqual(command.get_completion_snippet(),
            'foo | FOO_AUTO_COMPLETE')

    def test_handle(self):
        stream = io.StringIO()
        command = CompletionCommand('foo', stream)
        command.template = 'bar'
        command.handle(argparse.Namespace())
        self.assertEqual(stream.getvalue(), 'bar')

    def test_post_register(self):
        manager = ExecutionManager()
        with mock.patch.object(CompletionCommand, 'get_env_var_name') as m:
            m.return_value = 'FOO_AUTO_COMPLETE'
            manager.register('completion', CompletionCommand)
        self.assertTrue(manager.completion)
        self.assertEqual(manager.completion_env_var_name, 'FOO_AUTO_COMPLETE')


@mock.patch.object(os, 'environ', {'PROG_AUTO_COMPLETE': '1'})
class TestAutocomplete(unittest.TestCase):

    def setUp(self):
        self.stream = io.StringIO()
        self.stdout = io.StringIO()
        self.manager = ExecutionManager(stderr=self.stream, stdout=self.stdout)
        self.manager
        Command = type('Command', (BaseCommand,), {})
        CompCommand = type('CompCommand', (CompletionCommand,), {})

        class AddCommand(BaseCommand):
            args = BaseCommand.args + [
                arg('-f', '--force', action='store_true', default=False),
            ]

        self.manager.register('add', AddCommand)
        self.manager.register('annotate', Command)
        self.manager.register('init', Command)
        self.manager.register('completion', CompCommand)
        self.manager.completion = True
        self.manager.completion_env_var_name = 'PROG_AUTO_COMPLETE'

    @mock.patch.object(sys, 'argv', ['prog', 'a'])
    def test_autocomplete_returns_none_if_env_var_not_defined(self):
        with mock.patch.object(os, 'environ', {}):
            self.assertEqual(self.manager.autocomplete(), None)

    @mock.patch.object(sys, 'argv', ['prog', 'a'])
    def test_manager_execute_calls_autocomplete(self):
        self.manager.autocomplete = mock.Mock()
        stream = io.StringIO()
        with mock.patch.object(sys, 'stderr', stream):
            with self.assertRaises(SystemExit):
                self.manager.execute()
        self.manager.autocomplete.assert_called_once_with()

    @mock.patch.object(sys, 'argv', ['prog', 'a'])
    def test_autocomplete_returns_completes_for_subcommands(self):
        os.environ['COMP_WORDS'] = 'prog a'
        os.environ['COMP_CWORD'] = '1'
        stream = io.StringIO()
        with mock.patch.object(sys, 'stderr', stream):
            with self.assertRaises(SystemExit):
                self.manager.execute()
        self.assertEqual(self.stdout.getvalue(), 'add annotate')

    @mock.patch.object(sys, 'argv', ['prog', 'a'])
    def test_autocomplete_returns_none_if_wrong_COMP_CWORD_set(self):
        os.environ['COMP_WORDS'] = 'prog a'
        os.environ['COMP_CWORD'] = '3'
        stream = io.StringIO()
        with mock.patch.object(sys, 'stderr', stream):
            with self.assertRaises(SystemExit):
                self.manager.execute()
        self.assertEqual(self.stdout.getvalue(), '')

    # skip until we implement completion for subcommands' arguments
    #@mock.patch.object(sys, 'argv', ['prog', 'add', '--fo'])
    #def test_autocomplete_returns_completes_for_subcommands_args(self):
        #os.environ['COMP_WORDS'] = 'prog add --fo'
        #os.environ['COMP_CWORD'] = '2'
        #stream = io.StringIO()
        #with mock.patch.object(sys, 'stderr', stream):
            #with self.assertRaises(SystemExit):
                #self.manager.execute()
        #self.assertEqual(self.stdout.getvalue(), '--force')




import io
import sys
import mock
import argparse
from monolith.compat import unittest
from monolith.cli.base import arg
from monolith.cli.base import ExecutionManager
from monolith.cli.base import BaseCommand
from monolith.cli.base import LabelCommand
from monolith.cli.base import SingleLabelCommand
from monolith.cli.base import Parser
from monolith.cli.exceptions import AlreadyRegistered
from io import StringIO


class DummyCommand(BaseCommand):
    pass


class TestExecutionManager(unittest.TestCase):

    def assertRegistryClassesEqual(self, actual, expected):
        self.assertEqual(list(sorted(actual)), list(sorted(expected)))
        for key in actual:
            self.assertEqual(actual[key].__class__, expected[key],
                "Command class don't match for %r (it's %r but "
                "expected %r)" % (key, actual[key].__class__,
                expected[key]))

    def setUp(self):
        self.manager = ExecutionManager(['foobar'], stderr=StringIO())

    def test_init_prog_name(self):
        self.assertEqual(self.manager.prog_name, 'foobar')

    def test_init_stderr(self):
        manager = ExecutionManager()
        self.assertEqual(manager.stderr, sys.stderr)

    def test_default_argv(self):
        with mock.patch.object(sys, 'argv', ['vcs', 'foo', 'bar']):
            manager = ExecutionManager()
            self.assertEqual(manager.argv, ['foo', 'bar'])

    def test_get_usage(self):
        self.manager.usage = 'foobar baz'
        self.assertEqual(self.manager.get_usage(), 'foobar baz')

    def test_get_parser(self):
        self.manager.usage = 'foo bar'
        parser = self.manager.get_parser()
        self.assertIsInstance(parser, argparse.ArgumentParser)
        self.assertEqual(parser.prog, 'foobar') # argv[0]
        self.assertEqual(parser.usage, 'foo bar')
        self.assertEqual(parser.stream, self.manager.stderr)

    def test_register(self):
        Command = type('Command', (BaseCommand,), {})
        self.manager.register('foo', Command)
        self.assertRegistryClassesEqual(self.manager.registry, {'foo': Command})

    def test_register_raise_if_command_with_same_name_registered(self):
        Command = type('Command', (BaseCommand,), {})
        self.manager.register('foobar', Command)
        with self.assertRaises(AlreadyRegistered):
            self.manager.register('foobar', Command)

    def test_register_respects_force_argument(self):
        Command1 = type('Command', (BaseCommand,), {})
        Command2 = type('Command', (BaseCommand,), {})
        self.manager.register('foobar', Command1)
        self.manager.register('foobar', Command2, force=True)
        self.assertRegistryClassesEqual(self.manager.registry, {
            'foobar': Command2})

    def test_get_commands(self):
        FooCommand = type('FooCommand', (BaseCommand,), {})
        BarCommand = type('BarCommand', (BaseCommand,), {})
        self.manager.register('foo', FooCommand)
        self.manager.register('bar', BarCommand)
        self.assertEqual(list(self.manager.get_commands().keys()), ['bar', 'foo'])
        self.assertRegistryClassesEqual(self.manager.get_commands(), {
            'foo': FooCommand,
            'bar': BarCommand,
        })

    def test_get_commands_to_register(self):
        FooCommand = type('FooCommand', (BaseCommand,), {})
        BarCommand = type('BarCommand', (BaseCommand,), {})

        class Manager(ExecutionManager):

            def get_commands_to_register(self):
                return {
                    'foo': FooCommand,
                    'bar': BarCommand,
                }

        manager = Manager(['foobar'])
        self.assertRegistryClassesEqual(manager.registry, {
            'foo': FooCommand,
            'bar': BarCommand,
        })

    def test_call_command(self):

        class Command(BaseCommand):
            name = 'init'
            handle = mock.Mock()

        self.manager.register('init', Command)
        self.manager.call_command('init')
        self.assertTrue(Command.handle.called)

    def test_called_command_has_prog_name_properly_set(self):

        prog_names = []

        class Command(BaseCommand):
            name = 'init'
            def handle(self, namespace):
                prog_names.append(self.prog_name)

        self.manager.register('init', Command)
        self.manager.call_command('init')
        self.assertEqual(prog_names, ['foobar'])

    def test_call_command_with_args(self):

        class Command(BaseCommand):
            args = [
                arg('-f', '--force', action='store_true', default=False),
            ]
            name = 'add'
            handle = mock.Mock()

        self.manager.register('add', Command)
        self.manager.call_command('add', '-f')
        self.assertTrue(Command.handle.called)
        namespace = Command.handle.call_args[0][0]
        self.assertTrue(namespace.force)

    def test_execute_calls_handle_command(self):

        class Command(BaseCommand):
            args = [
                arg('-f', '--force', action='store_true', default=False),
            ]
            name = 'add'
            handle = mock.Mock()

        self.manager.register('add', Command)
        with mock.patch.object(sys, 'argv', ['prog', 'add', '-f']):
            self.manager.execute()
        namespace = Command.handle.call_args[0][0]
        Command.handle.assert_called_once_with(namespace)


class TestBaseCommand(unittest.TestCase):

    def test_get_args(self):
        Command = type('Command', (BaseCommand,), {'args': ['foo', 'bar']})
        command = Command()
        self.assertEqual(command.get_args(), ['foo', 'bar'])

    def test_handle_raises_error(self):
        with self.assertRaises(NotImplementedError):
            BaseCommand().handle(argparse.Namespace())

    def test_post_register_hooks(self):
        Command = type('Command', (BaseCommand,), {'args': ['foo', 'bar']})

        class Command(BaseCommand):
            def post_register(self, manager):
                manager.completion = True

        manager = ExecutionManager()
        self.assertFalse(manager.completion)
        manager.register('completion', Command)
        self.assertTrue(manager.completion)


class TestLabelCommand(unittest.TestCase):

    def test_handle_raise_if_handle_label_not_implemented(self):
        command = LabelCommand()
        with self.assertRaises(NotImplementedError):
            command.handle(argparse.Namespace(labels=['foo']))

    def test_handle_calls_handle_label(self):
        namespace = argparse.Namespace(labels=['foo', 'bar'])
        command = LabelCommand()
        command.handle_label = mock.Mock()
        command.handle(namespace)
        self.assertEqual(command.handle_label.call_args_list, [
            arg('foo', namespace),
            arg('bar', namespace),
        ])

    def test_labels_required_true(self):
        Command = type('Command', (LabelCommand,), {'labels_required': True})
        command = Command()
        self.assertEqual(command.get_args()[0].kwargs.get('nargs'), '+')

    def test_labels_required_false(self):
        Command = type('Command', (LabelCommand,), {'labels_required': False})
        command = Command()
        self.assertEqual(command.get_args()[0].kwargs.get('nargs'), '*')

    def test_handle_no_labels_called_if_no_labels_given(self):
        Command = type('Command', (LabelCommand,), {'labels_required': False})
        command = Command()
        command.handle_no_labels = mock.Mock()
        namespace = argparse.Namespace(labels=[])
        command.handle(namespace)
        command.handle_no_labels.assert_called_once_with(namespace)


class TestSingleLabelCommand(unittest.TestCase):

    def test_get_label_arg(self):
        Command = type('Command', (SingleLabelCommand,), {})
        label_arg = Command().get_label_arg()
        self.assertEqual(label_arg, arg('label',
            default=Command.label_default_value, nargs='?'))

    def test_get_args(self):
        Command = type('Command', (SingleLabelCommand,), {})
        command = Command()
        self.assertEqual(command.get_args(), [command.get_label_arg()])

    def test_handle_raise_if_handle_label_not_implemented(self):
        command = SingleLabelCommand()
        with self.assertRaises(NotImplementedError):
            command.handle(argparse.Namespace(label='foo'))

    def test_handle_calls_handle_label(self):
        namespace = argparse.Namespace(label='foobar')
        command = SingleLabelCommand()
        command.handle_label = mock.Mock()
        command.handle(namespace)
        self.assertEqual(command.handle_label.call_args_list, [
            arg('foobar', namespace),
        ])


class TestArg(unittest.TestCase):

    def test_args(self):
        self.assertEqual(arg(1, 2, 'foo', bar='baz').args, (1, 2, 'foo'))

    def test_kargs(self):
        self.assertEqual(arg(1, 2, 'foo', bar='baz').kwargs, {'bar': 'baz'})


class TestParser(unittest.TestCase):

    def setUp(self):
        self.stream = io.StringIO()
        self.parser = Parser(stream=self.stream)

    def test_print_message_default_file(self):
        self.parser._print_message('foobar')
        self.assertEqual(self.stream.getvalue(), 'foobar')

